from functools import partial
from types import ModuleType
from typing import Literal, TypeAlias
import itertools

import numpy as np

from lucid._tensor import Tensor
from lucid._backend.core import (
    Operation,
    func_op,
    _FuncOpReturnType,
    _GradType,
)
from lucid._backend.metal import mx
from lucid.types import _NumPyArray, _MLXArray


_Array: TypeAlias = _NumPyArray | _MLXArray
_Shape: TypeAlias = tuple[int, ...]
_Mode: TypeAlias = Literal["avg", "max"]


def _to_tuple(value: int | tuple[int, ...] | list[int], dim: int, name: str) -> _Shape:
    if isinstance(value, int):
        return (value,) * dim

    if isinstance(value, (tuple, list)):
        if len(value) == 1:
            return (int(value[0]),) * dim
        if len(value) != dim:
            raise ValueError(f"{name} must have length {dim}, got {len(value)}.")
        return tuple(int(v) for v in value)

    raise TypeError(f"{name} must be int or sequence, got {type(value).__name__}.")


def _prod(shape: _Shape) -> int:
    total = 1
    for v in shape:
        total *= int(v)
    return total


def _pool_out_dims(
    input_spatial: _Shape, kernel_size: _Shape, stride: _Shape, padding: _Shape
) -> _Shape:
    out_dims = []
    for i in range(len(kernel_size)):
        o = (input_spatial[i] + 2 * padding[i] - kernel_size[i]) // stride[i] + 1
        if o <= 0:
            raise ValueError(f"Non-positive output dim for axis {i}: {o}")
        out_dims.append(o)
    return tuple(out_dims)


def _pad_input(lib_: ModuleType, data: _Array, padding: _Shape) -> _Array:
    if not any(padding):
        return data
    pad_width = ((0, 0), (0, 0)) + tuple((p, p) for p in padding)
    return lib_.pad(data, pad_width)


def _crop_padding(data: _Array, padding: _Shape) -> _Array:
    if not any(padding):
        return data

    slices = [slice(None), slice(None)]
    for p in padding:
        end = -p if p != 0 else None
        slices.append(slice(p, end))
    return data[tuple(slices)]


def _zeros(lib_: ModuleType, shape: _Shape, dtype: object) -> _Array:
    if lib_ is np:
        return np.zeros(shape, dtype=dtype)
    return mx.zeros(shape, dtype=dtype)


def _full_like_int(lib_: ModuleType, ref: _Array, value: int) -> _Array:
    shape = ref.shape
    if lib_ is np:
        return np.full(shape, value, dtype=np.int32)
    return mx.full(shape, value, dtype=mx.int32)


def _where(lib_: ModuleType, cond: _Array, x: _Array, y: _Array) -> _Array:
    if lib_ is np:
        return np.where(cond, x, y)
    return mx.where(cond, x, y)


def _pool_forward_sum(
    x_pad: _Array, out_dims: _Shape, kernel_size: _Shape, stride: _Shape
) -> _Array:
    out = None
    for k_idx in itertools.product(*[range(k) for k in kernel_size]):
        slices = [slice(None), slice(None)]
        for d in range(len(kernel_size)):
            start = k_idx[d]
            end = start + stride[d] * out_dims[d]
            slices.append(slice(start, end, stride[d]))

        window = x_pad[tuple(slices)]
        out = window if out is None else out + window

    return out


def _pool_forward_max(
    lib_: ModuleType,
    x_pad: _Array,
    out_dims: _Shape,
    kernel_size: _Shape,
    stride: _Shape,
) -> tuple[_Array, _Array]:
    max_vals = None
    max_idx = None

    for flat_idx, k_idx in enumerate(
        itertools.product(*[range(k) for k in kernel_size])
    ):
        slices = [slice(None), slice(None)]
        for d in range(len(kernel_size)):
            start = k_idx[d]
            end = start + stride[d] * out_dims[d]
            slices.append(slice(start, end, stride[d]))

        window = x_pad[tuple(slices)]
        if max_vals is None:
            max_vals = window
            max_idx = _full_like_int(lib_, window, flat_idx)
            continue

        mask = window > max_vals
        max_vals = _where(lib_, mask, window, max_vals)
        idx_arr = _full_like_int(lib_, max_vals, flat_idx)
        max_idx = _where(lib_, mask, idx_arr, max_idx)

    return max_vals, max_idx


def _pool_backward_avg(
    lib_: ModuleType,
    grad_out: _Array,
    input_shape: _Shape,
    out_dims: _Shape,
    kernel_size: _Shape,
    stride: _Shape,
    padding: _Shape,
) -> _Array:
    pad_shape = list(input_shape)
    for i, p in enumerate(padding):
        pad_shape[2 + i] += 2 * p

    grad_input_pad = _zeros(lib_, tuple(pad_shape), dtype=grad_out.dtype)
    grad_scaled = grad_out / _prod(kernel_size)

    for k_idx in itertools.product(*[range(k) for k in kernel_size]):
        slices = [slice(None), slice(None)]
        for d in range(len(kernel_size)):
            start = k_idx[d]
            end = start + stride[d] * out_dims[d]
            slices.append(slice(start, end, stride[d]))

        if lib_ is np:
            grad_input_pad[tuple(slices)] += grad_scaled
        else:
            grad_input_pad = grad_input_pad.at[tuple(slices)].add(grad_scaled)

    return _crop_padding(grad_input_pad, padding)


def _pool_backward_max(
    lib_: ModuleType,
    grad_out: _Array,
    input_shape: _Shape,
    out_dims: _Shape,
    kernel_size: _Shape,
    stride: _Shape,
    padding: _Shape,
    max_idx: _Array,
) -> _Array:
    pad_shape = list(input_shape)
    for i, p in enumerate(padding):
        pad_shape[2 + i] += 2 * p

    grad_input_pad = _zeros(lib_, tuple(pad_shape), dtype=grad_out.dtype)

    for flat_idx, k_idx in enumerate(
        itertools.product(*[range(k) for k in kernel_size])
    ):
        slices = [slice(None), slice(None)]
        for d in range(len(kernel_size)):
            start = k_idx[d]
            end = start + stride[d] * out_dims[d]
            slices.append(slice(start, end, stride[d]))

        mask = max_idx == flat_idx
        grad_contrib = grad_out * mask
        if lib_ is np:
            grad_input_pad[tuple(slices)] += grad_contrib
        else:
            grad_input_pad = grad_input_pad.at[tuple(slices)].add(grad_contrib)

    return _crop_padding(grad_input_pad, padding)


class pool_nd_kernel(Operation):
    def __init__(
        self,
        kernel_size: int | tuple[int, ...] | list[int],
        stride: int | tuple[int, ...] | list[int],
        padding: int | tuple[int, ...] | list[int],
        mode: _Mode,
    ) -> None:
        super().__init__()
        if mode not in {"avg", "max"}:
            raise ValueError(f"Unsupported pooling mode: {mode}")

        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.mode = mode

        self._kernel_size: _Shape | None = None
        self._stride: _Shape | None = None
        self._padding: _Shape | None = None
        self._out_dims: _Shape | None = None
        self._input_shape: _Shape | None = None
        self._max_idx: _Array | None = None

    def clear(self) -> None:
        self.result = None
        self._kernel_size = None
        self._stride = None
        self._padding = None
        self._out_dims = None
        self._input_shape = None
        self._max_idx = None

    def _normalize(self, input_: Tensor) -> tuple[_Shape, _Shape, _Shape]:
        if input_.ndim < 3:
            raise ValueError("Input must have at least 3 dimensions (N, C, ...).")

        D = input_.ndim - 2
        kernel = _to_tuple(self.kernel_size, D, "kernel_size")
        stride = _to_tuple(self.stride, D, "stride")
        padding = _to_tuple(self.padding, D, "padding")

        self._kernel_size = kernel
        self._stride = stride
        self._padding = padding

        return kernel, stride, padding

    @func_op(n_in=1, n_ret=1)
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        kernel, stride, padding = self._normalize(a)
        out_dims = _pool_out_dims(a.shape[2:], kernel, stride, padding)
        self._out_dims = out_dims
        self._input_shape = a.shape

        x_pad = _pad_input(np, a.data, padding)
        if self.mode == "avg":
            out_sum = _pool_forward_sum(x_pad, out_dims, kernel, stride)
            out = out_sum / _prod(kernel)
        else:
            out, max_idx = _pool_forward_max(np, x_pad, out_dims, kernel, stride)
            self._max_idx = max_idx

        self.result = Tensor(out)
        return self.result, partial(self.__grad__, lib_=np)

    @func_op(n_in=1, n_ret=1, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        kernel, stride, padding = self._normalize(a)
        out_dims = _pool_out_dims(a.shape[2:], kernel, stride, padding)
        self._out_dims = out_dims
        self._input_shape = a.shape

        x_pad = _pad_input(mx, a.data, padding)
        if self.mode == "avg":
            out_sum = _pool_forward_sum(x_pad, out_dims, kernel, stride)
            out = out_sum / _prod(kernel)
        else:
            out, max_idx = _pool_forward_max(mx, x_pad, out_dims, kernel, stride)
            self._max_idx = max_idx

        self.result = Tensor(out)
        return self.result, partial(self.__grad__, lib_=mx)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if (
            self._kernel_size is None
            or self._stride is None
            or self._padding is None
            or self._out_dims is None
            or self._input_shape is None
        ):
            raise RuntimeError("pool_nd backward called before forward.")

        grad_out = self.result.grad
        if self.mode == "avg":
            grad_input = _pool_backward_avg(
                lib_,
                grad_out,
                self._input_shape,
                self._out_dims,
                self._kernel_size,
                self._stride,
                self._padding,
            )
            return grad_input

        if self._max_idx is None:
            raise RuntimeError("max pool backward missing max indices.")

        grad_input = _pool_backward_max(
            lib_,
            grad_out,
            self._input_shape,
            self._out_dims,
            self._kernel_size,
            self._stride,
            self._padding,
            self._max_idx,
        )
        return grad_input

    def __flops__(self, a: Tensor) -> int:
        if self._kernel_size is None or self._out_dims is None:
            kernel, stride, padding = self._normalize(a)
            out_dims = _pool_out_dims(a.shape[2:], kernel, stride, padding)
        else:
            kernel = self._kernel_size
            out_dims = self._out_dims

        kernel_elems = _prod(kernel)
        out_elems = int(a.shape[0]) * int(a.shape[1]) * _prod(out_dims)

        if kernel_elems <= 0 or out_elems <= 0:
            return 0

        if self.mode == "avg":
            return out_elems * kernel_elems
        return out_elems * max(kernel_elems - 1, 0)



import math
from typing import Literal

import lucid
import lucid.nn as nn
import lucid.nn.functional as F

from lucid._tensor import Tensor
from lucid import register_model
from lucid.types import _Scalar, _ShapeLike, _DeviceType

import lucid.models.imgclf.efficient as effnet
from lucid.models.objdet.util import iou, nms, DetectionDict


__all__ = [
    "EfficientDet",
    "efficientdet_d0",
    "efficientdet_d1",
    "efficientdet_d2",
    "efficientdet_d3",
    "efficientdet_d4",
    "efficientdet_d5",
    "efficientdet_d6",
    "efficientdet_d7",
]


_CompoundCoefLiteral = Literal[0, 1, 2, 3, 4, 5, 6, 7]


class _ConvBlock(nn.Module):
    def __init__(self, num_channels: int) -> None:
        super().__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(
                num_channels,
                num_channels,
                kernel_size=3,
                padding=1,
                groups=num_channels,
            ),
            nn.Conv2d(num_channels, num_channels, kernel_size=1),
            nn.BatchNorm2d(num_channels, momentum=0.9997, eps=4e-5),
            nn.ReLU(),
        )

    def forward(self, x: Tensor) -> Tensor:
        return self.conv(x)


class _BiFPN(nn.Module):
    def __init__(self, num_channels: int, eps: float = 1e-4) -> None:
        super().__init__()
        self.eps = eps
        self.convs = nn.ModuleDict(
            {
                **{f"{i}_up": _ConvBlock(num_channels) for i in range(3, 7)},
                **{f"{i}_down": _ConvBlock(num_channels) for i in range(4, 8)},
            }
        )
        self.ups = nn.ModuleDict(
            {f"{i}": nn.Upsample(scale_factor=2, mode="nearest") for i in range(3, 7)}
        )
        self.downs = nn.ModuleDict(
            {f"{i}": nn.AvgPool2d(kernel_size=2, stride=2) for i in range(4, 8)}
        )

        self.weights = nn.ParameterDict(
            {
                **{f"{i}_w1": nn.Parameter(lucid.ones(2)) for i in range(3, 7)},
                **{f"{i}_w2": nn.Parameter(lucid.ones(3)) for i in range(4, 8)},
            }
        )
        self.acts = nn.ModuleDict(
            {
                **{f"{i}_w1": nn.Swish() for i in range(3, 7)},
                **{f"{i}_w2": nn.Swish() for i in range(4, 8)},
            }
        )

    def _norm_weight(self, weight: Tensor) -> Tensor:
        return weight / (weight.sum(axis=0) + self.eps)

    def _forward_up(self, feats: tuple[Tensor]) -> tuple[Tensor]:
        p3_in, p4_in, p5_in, p6_in, p7_in = feats

        w1_p6_up = self._norm_weight(self.acts["6_w1"](self.weights["6_w1"]))
        p6_up_in = w1_p6_up[0] * p6_in + w1_p6_up[1] * self.ups["6"](p7_in)
        p6_up = self.convs["6_up"](p6_up_in)

        w1_p5_up = self._norm_weight(self.acts["5_w1"](self.weights["5_w1"]))
        p5_up_in = w1_p5_up[0] * p5_in + w1_p5_up[1] * self.ups["5"](p6_up)
        p5_up = self.convs["5_up"](p5_up_in)

        w1_p4_up = self._norm_weight(self.acts["4_w1"](self.weights["4_w1"]))
        p4_up_in = w1_p4_up[0] * p4_in + w1_p4_up[1] * self.ups["4"](p5_up)
        p4_up = self.convs["4_up"](p4_up_in)

        w1_p3_up = self._norm_weight(self.acts["3_w1"](self.weights["3_w1"]))
        p3_up_in = w1_p3_up[0] * p3_in + w1_p3_up[1] * self.ups["3"](p4_up)
        p3_out = self.convs["3_up"](p3_up_in)

        return p3_out, p4_up, p5_up, p6_up

    def _forward_down(
        self, feats: tuple[Tensor], up_feats: tuple[Tensor]
    ) -> tuple[Tensor]:
        _, p4_in, p5_in, p6_in, p7_in = feats
        p3_out, p4_up, p5_up, p6_up = up_feats

        w2_p4_down = self._norm_weight(self.acts["4_w2"](self.weights["4_w2"]))
        p4_down_in = (
            w2_p4_down[0] * p4_in
            + w2_p4_down[1] * p4_up
            + w2_p4_down[2] * self.downs["4"](p3_out)
        )
        p4_out = self.convs["4_down"](p4_down_in)

        w2_p5_down = self._norm_weight(self.acts["5_w2"](self.weights["5_w2"]))
        p5_down_in = (
            w2_p5_down[0] * p5_in
            + w2_p5_down[1] * p5_up
            + w2_p5_down[2] * self.downs["5"](p4_out)
        )
        p5_out = self.convs["5_down"](p5_down_in)

        w2_p6_down = self._norm_weight(self.acts["6_w2"](self.weights["6_w2"]))
        p6_down_in = (
            w2_p6_down[0] * p6_in
            + w2_p6_down[1] * p6_up
            + w2_p6_down[2] * self.downs["6"](p5_out)
        )
        p6_out = self.convs["6_down"](p6_down_in)

        w2_p7_down = self._norm_weight(self.acts["7_w2"](self.weights["7_w2"]))
        p7_down_in = w2_p7_down[0] * p7_in + w2_p7_down[1] * self.downs["7"](p6_out)
        p7_out = self.convs["7_down"](p7_down_in)

        return p3_out, p4_out, p5_out, p6_out, p7_out

    def forward(self, feats: tuple[Tensor, ...]) -> tuple[Tensor, ...]:
        up_feats = self._forward_up(feats)
        down_feats = self._forward_down(feats, up_feats)

        return down_feats


class _BBoxRegressor(nn.Module):
    def __init__(self, in_channels: int, num_anchors: int, num_layers: int) -> None:
        super().__init__()
        layers: list[nn.Module] = []
        for _ in range(num_layers):
            layers.extend(
                [
                    nn.DepthSeparableConv2d(
                        in_channels, in_channels, kernel_size=3, padding=1
                    ),
                    nn.BatchNorm2d(in_channels, momentum=0.99, eps=1e-3),
                    nn.Swish(),
                ]
            )

        self.layers = nn.Sequential(*layers)
        self.header = nn.Conv2d(
            in_channels,
            num_anchors * 4,
            kernel_size=3,
            padding=1,
        )

    def forward(self, x: Tensor) -> Tensor:
        x = self.layers(x)
        x = self.header(x)

        x = x.transpose((0, 2, 3, 1))
        return x.reshape(x.shape[0], -1, 4)


class _Classifier(nn.Module):
    def __init__(
        self, in_channels: int, num_anchors: int, num_classes: int, num_layers: int
    ) -> None:
        super().__init__()
        self.num_anchors = num_anchors
        self.num_classes = num_classes

        layers: list[nn.Module] = []
        for _ in range(num_layers):
            layers.extend(
                [
                    nn.DepthSeparableConv2d(
                        in_channels, in_channels, kernel_size=3, padding=1
                    ),
                    nn.BatchNorm2d(in_channels, momentum=0.99, eps=1e-3),
                    nn.Swish(),
                ]
            )

        self.layers = nn.Sequential(*layers)
        self.header = nn.Conv2d(
            in_channels,
            num_anchors * num_classes,
            kernel_size=3,
            padding=1,
        )
        self.act = nn.Sigmoid()

    def forward(self, x: Tensor) -> Tensor:
        x = self.layers(x)
        x = self.header(x)
        x = self.act(x)

        x = x.transpose((0, 2, 3, 1))
        return x.reshape(x.shape[0], -1, self.num_classes)


def _generate_anchors(
    base_size: int, ratios: list[_Scalar], scales: list[_Scalar], device: _DeviceType
) -> Tensor:
    num_anchors = len(ratios) * len(scales)
    anchors = lucid.zeros(num_anchors, 4)
    anchors[:, 2:] = base_size * lucid.tile(scales, reps=(2, len(ratios))).T

    areas = anchors[:, 2] * anchors[:, 3]

    anchors[:, 2] = lucid.sqrt(areas / lucid.repeat(ratios, len(scales)))
    anchors[:, 3] = anchors[:, 2] * lucid.repeat(ratios, len(scales))

    anchors[:, 0::2] -= lucid.tile(anchors[:, 2] * 0.5, reps=(2, 1)).T
    anchors[:, 1::2] -= lucid.tile(anchors[:, 3] * 0.5, reps=(2, 1)).T

    return anchors.to(device)


def _shift_anchors(shape: _ShapeLike, stride: int, anchors: Tensor) -> Tensor:
    shift_x = (lucid.arange(0, shape[1], device=anchors.device) + 0.5) * stride
    shift_y = (lucid.arange(0, shape[0], device=anchors.device) + 0.5) * stride

    shift_x, shift_y = lucid.meshgrid(shift_x, shift_y, indexing="xy")
    shifts = lucid.vstack(
        (shift_x.ravel(), shift_y.ravel(), shift_x.ravel(), shift_y.ravel())
    ).T

    A = anchors.shape[0]
    K = shifts.shape[0]

    all_anchors = anchors.reshape(1, A, 4) + shifts.reshape(K, 1, 4)
    all_anchors = all_anchors.reshape(K * A, 4)

    return all_anchors


@nn.auto_repr("pyramid_levels")
class _Anchors(nn.Module):
    def __init__(
        self,
        pyramid_levels: list[int] | None = None,
        strides: list[int] | None = None,
        sizes: list[int] | None = None,
        ratios: list[int] | None = None,
        scales: list[int] | None = None,
    ) -> None:
        super().__init__()
        self.pyramid_levels = (
            [3, 4, 5, 6, 7] if pyramid_levels is None else pyramid_levels
        )
        self.strides = (
            [2**x for x in self.pyramid_levels] if strides is None else strides
        )
        self.sizes = (
            [2 ** (x + 2) for x in self.pyramid_levels] if sizes is None else sizes
        )
        self.ratios = [0.5, 1.0, 2.0] if ratios is None else ratios
        self.scales = [2**i for i in (0, 1 / 3, 2 / 3)] if scales is None else scales

    def forward(self, x: Tensor) -> Tensor:
        img_shape = Tensor(x.shape[2:], dtype=lucid.Int16, device=x.device)
        img_shapes: list[Tensor] = [
            (img_shape + 2**i - 1) // (2**i) for i in self.pyramid_levels
        ]

        anchors_per_level: list[Tensor] = []
        for idx, _ in enumerate(self.pyramid_levels):
            anchors = _generate_anchors(
                self.sizes[idx], self.ratios, self.scales, device=x.device
            )
            shifted_anchors = _shift_anchors(
                img_shapes[idx].tolist(), self.strides[idx], anchors
            )
            anchors_per_level.append(shifted_anchors)

        all_anchors = lucid.concatenate(anchors_per_level, axis=0)
        all_anchors = all_anchors.unsqueeze(axis=0)

        return all_anchors


class _BBoxTransform(nn.Module):
    def __init__(self, mean: Tensor | None = None, std: Tensor | None = None) -> None:
        super().__init__()
        self.mean = lucid.zeros(4) if mean is None else mean
        self.std = Tensor([0.1, 0.1, 0.2, 0.2]) if std is None else std

    def forward(self, boxes: Tensor, deltas: Tensor) -> Tensor:
        widths = boxes[:, :, 2] - boxes[:, :, 0]
        heights = boxes[:, :, 3] - boxes[:, :, 1]

        ctr_x = boxes[:, :, 0] + 0.5 * widths
        ctr_y = boxes[:, :, 1] + 0.5 * heights

        def _inv_norm(d: Tensor, s: Tensor, m: Tensor) -> Tensor:
            return d * s + m

        dx, dy, dw, dh = (
            _inv_norm(deltas[:, :, i], self.std[i], self.mean[i]) for i in range(4)
        )

        pred_ctr_x = ctr_x + dx * widths
        pred_ctr_y = ctr_y + dy * heights
        pred_w = lucid.exp(dw) * widths
        pred_h = lucid.exp(dh) * heights

        pred_box_x1 = pred_ctr_x - 0.5 * pred_w
        pred_box_y1 = pred_ctr_y - 0.5 * pred_h
        pred_box_x2 = pred_ctr_x + 0.5 * pred_w
        pred_box_y2 = pred_ctr_y + 0.5 * pred_h

        pred_boxes = lucid.stack(
            [pred_box_x1, pred_box_y1, pred_box_x2, pred_box_y2], axis=2
        )
        return pred_boxes


class _ClipBoxes(nn.Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, boxes: Tensor, images: Tensor) -> Tensor:
        boxes[:, :, 0] = lucid.clip(boxes[:, :, 0], min_value=0)
        boxes[:, :, 1] = lucid.clip(boxes[:, :, 1], min_value=0)

        boxes[:, :, 2] = lucid.clip(boxes[:, :, 2], min_value=images.shape[3])
        boxes[:, :, 3] = lucid.clip(boxes[:, :, 3], min_value=images.shape[2])

        return boxes


class _FocalLoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.alpha = 0.25
        self.gamma = 2.0

    def forward(
        self, clfs: Tensor, regs: Tensor, anchors: Tensor, annotations: Tensor
    ) -> tuple[Tensor, Tensor]:
        N = clfs.shape[0]
        device = clfs.device
        clf_losses = []
        reg_losses = []

        anchor = anchors[0, :, :]
        anchor_w = anchor[:, 2] - anchor[:, 0]
        anchor_h = anchor[:, 3] - anchor[:, 1]

        anchor_ctr_x = anchor[:, 0] + 0.5 * anchor_w
        anchor_ctr_y = anchor[:, 1] + 0.5 * anchor_h

        for i in range(N):
            clf = clfs[i, :, :]
            reg = regs[i, :, :]

            box_annot = annotations[i, :, :]
            box_annot = box_annot[box_annot[:, 4] != 1]

            if len(box_annot) == 0:
                reg_losses.append(lucid.zeros(1, dtype=lucid.Float32, device=device))
                clf_losses.append(lucid.zeros(1, dtype=lucid.Float32, device=device))
                continue

            clf = lucid.clip(clf, 1e-4, 1.0 - 1e-4)

            iou_val = iou(anchor, box_annot[:, :4])
            iou_max = lucid.max(iou_val, axis=1)
            iou_argmax = lucid.argmax(iou_val, axis=1)

            targets = lucid.ones(*clf.shape, device=device) * -1
            targets[iou_max < 0.4, :] = 0

            pos_indices = iou_max >= 0.5
            num_pos_anchors = pos_indices.sum()
            assigned_annot = box_annot[iou_argmax, :]

            targets[pos_indices, :] = 0
            targets[pos_indices, assigned_annot[pos_indices, 4].astype(lucid.Int)] = 1

            alpha_factor = lucid.ones(*targets.shape, device=device) * self.alpha
            alpha_factor = lucid.where(targets == 1.0, alpha_factor, 1.0 - alpha_factor)

            focal_weight = lucid.where(targets == 1.0, 1.0 - clf, clf)
            focal_weight = alpha_factor * focal_weight**self.gamma

            bce = F.binary_cross_entropy(clf, targets, reduction=None)

            cls_loss = focal_weight * bce
            cls_loss = lucid.where(
                targets != -1.0,
                cls_loss,
                lucid.zeros(*cls_loss.shape, device=device),
            )

            clf_losses.append(
                cls_loss.sum() / lucid.clip(num_pos_anchors, min_value=1.0)
            )

            if pos_indices.sum() > 0:
                assigned_annot = assigned_annot[pos_indices, :]

                anchor_w_pi = anchor_w[pos_indices]
                anchor_h_pi = anchor_h[pos_indices]

                anchor_ctr_x_pi = anchor_ctr_x[pos_indices]
                anchor_ctr_y_pi = anchor_ctr_y[pos_indices]

                gt_widths = assigned_annot[:, 2] - assigned_annot[:, 0]
                gt_heights = assigned_annot[:, 3] - assigned_annot[:, 1]

                gt_ctr_x = assigned_annot[:, 0] + 0.5 * gt_widths
                gt_ctr_y = assigned_annot[:, 1] + 0.5 * gt_heights

                gt_widths = lucid.clip(gt_widths, min_value=1)
                gt_heights = lucid.clip(gt_heights, min_value=1)

                targets_dx = (gt_ctr_x - anchor_ctr_x_pi) / anchor_w_pi
                targets_dy = (gt_ctr_y - anchor_ctr_y_pi) / anchor_h_pi
                targets_dw = lucid.log(gt_widths / anchor_w_pi)
                targets_dh = lucid.log(gt_heights / anchor_h_pi)

                targets = lucid.stack(
                    [targets_dx, targets_dy, targets_dw, targets_dh]
                ).T / Tensor([0.1, 0.1, 0.2, 0.2])

                reg_diff = lucid.abs(targets - reg[pos_indices, :])
                reg_loss = lucid.where(
                    reg_diff <= 1 / 9, 0.5 * 9.0 * reg_diff**2, reg_diff - 0.5 / 9.0
                )
                reg_losses.append(reg_loss.mean())

            else:
                reg_losses.append(lucid.zeros(1, dtype=lucid.Float32, device=device))

        return (
            lucid.stack(clf_losses).mean(axis=0, keepdims=True),
            lucid.stack(reg_losses).mean(axis=0, keepdims=True),
        )


class _EfficientNetBackbone(nn.Module):
    def __init__(self, compound_coef: _CompoundCoefLiteral) -> None:
        super().__init__()
        model = getattr(effnet, f"efficientnet_b{compound_coef}")()
        self.model = nn.Sequential(*[getattr(model, f"stage{i}") for i in range(1, 8)])

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor, Tensor]:
        out: list[Tensor] = []
        for idx, stage in enumerate(self.model):
            x = stage(x)
            if idx in {3, 4, 6}:
                out.append(x)

        return tuple(out)


_backbone_channels: dict[int, tuple[int]] = {
    0: (40, 80, 192),
    1: (40, 80, 192),
    2: (48, 88, 208),
    3: (48, 88, 208),
    4: (56, 112, 224),
    5: (64, 160, 288),
    6: (72, 192, 384),
    7: (72, 192, 384),
}


class EfficientDet(nn.Module):
    def __init__(
        self,
        compound_coef: _CompoundCoefLiteral = 0,
        num_anchors: int = 9,
        num_classes: int = 80,
    ) -> None:
        super().__init__()
        self.num_classes = num_classes
        self.compound_coef = compound_coef

        base_channels = 64
        self.num_channels = int(round(base_channels * (1.4**self.compound_coef)))
        self.num_channels = int(math.ceil(self.num_channels / 8) * 8)

        c3_in, c4_in, c5_in = _backbone_channels[self.compound_coef]

        self.conv3 = nn.Conv2d(c3_in, self.num_channels, kernel_size=1, padding=0)
        self.conv4 = nn.Conv2d(c4_in, self.num_channels, kernel_size=1, padding=0)
        self.conv5 = nn.Conv2d(c5_in, self.num_channels, kernel_size=1, padding=0)
        self.conv6 = nn.Conv2d(
            c5_in, self.num_channels, kernel_size=3, stride=2, padding=1
        )
        self.conv7 = nn.Sequential(
            nn.ReLU(),
            nn.Conv2d(
                self.num_channels, self.num_channels, kernel_size=3, stride=2, padding=1
            ),
        )

        self.bifpn = nn.Sequential(
            *[_BiFPN(self.num_channels) for _ in range(min(2 + self.compound_coef, 8))]
        )
        self.regressor = _BBoxRegressor(
            self.num_channels, num_anchors, num_layers=3 + self.compound_coef // 3
        )
        self.classifier = _Classifier(
            self.num_channels,
            num_anchors,
            num_classes,
            num_layers=3 + self.compound_coef // 3,
        )

        self.anchors = _Anchors()
        self.boxtrans = _BBoxTransform()
        self.clipbox = _ClipBoxes()
        self.loss = _FocalLoss()

        self.apply(self.init_weights)

        prior = 0.01
        self._header_weights_post_init(
            self.classifier.header, w_val=0.0, b_val=-math.log((1.0 - prior) / prior)
        )
        self._header_weights_post_init(self.regressor.header, w_val=0.0, b_val=0.0)

        self.backbone = _EfficientNetBackbone(self.compound_coef)

    def init_weights(self, m: nn.Module) -> None:
        if isinstance(m, nn.Conv2d):
            n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
            nn.init.normal(m.weight, 0.0, math.sqrt(2 / n))

        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant(m.weight, 1.0)
            nn.init.constant(m.bias, 0.0)

    @staticmethod
    def _header_weights_post_init(m: nn.Module, w_val: float, b_val: float) -> None:
        assert isinstance(m, nn.Conv2d)
        nn.init.constant(m.weight, w_val)
        nn.init.constant(m.bias, b_val)

    def forward(self, x: Tensor) -> tuple[Tensor, Tensor, Tensor]:
        c3, c4, c5 = self.backbone(x)

        p3 = self.conv3(c3)
        p4 = self.conv4(c4)
        p5 = self.conv5(c5)
        p6 = self.conv6(c5)
        p7 = self.conv7(p6)

        feats = (p3, p4, p5, p6, p7)
        feats = self.bifpn(feats)

        cls_preds = [self.classifier(f) for f in feats]
        box_preds = [self.regressor(f) for f in feats]

        cls_preds = lucid.concatenate(cls_preds, axis=1)
        box_preds = lucid.concatenate(box_preds, axis=1)
        anchors = self.anchors(x)

        return cls_preds, box_preds, anchors

    @lucid.no_grad()
    def predict(self, x: Tensor) -> list[list[DetectionDict]]:
        cls_preds, box_preds, anchors = self.forward(x)

        boxes = self.boxtrans(anchors, box_preds)
        boxes = self.clipbox(boxes, x)

        B = x.shape[0]
        results: list[list[DetectionDict]] = []

        for i in range(B):
            cls_i = cls_preds[i]
            box_i = boxes[i]

            scores = lucid.max(cls_i, axis=1)
            labels = lucid.argmax(cls_i, axis=1)

            keep = scores > 0.05
            scores, labels, box_i = scores[keep], labels[keep], box_i[keep]

            if scores.size == 0:
                results.append([])
                continue

            keep_idx = nms(box_i, scores, iou_thresh=0.3)
            scores, labels, box_i = scores[keep_idx], labels[keep_idx], box_i[keep_idx]

            dets: list[DetectionDict] = []
            for s, l, b in zip(scores, labels, box_i):
                dets.append(
                    {"box": b.tolist(), "score": s.item(), "class_id": int(l.item())}
                )
            results.append(dets)

        return results

    def get_loss(self, x: Tensor, targets: list[Tensor]) -> Tensor:
        cls_preds, box_preds, anchors = self.forward(x)

        max_n = max(t.shape[0] for t in targets)
        padded = lucid.zeros(
            (len(targets), max_n, 5), dtype=lucid.Float32, device=x.device
        )
        for i, t in enumerate(targets):
            padded[i, : t.shape[0]] = t

        cls_loss, reg_loss = self.loss(cls_preds, box_preds, anchors, padded)
        total_loss = cls_loss + reg_loss

        return total_loss


@register_model
def efficientdet_d0(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=0, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d1(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=1, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d2(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=2, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d3(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=3, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d4(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=4, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d5(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=5, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d6(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=6, num_classes=num_classes, **kwargs)


@register_model
def efficientdet_d7(num_classes: int = 80, **kwargs) -> EfficientDet:
    return EfficientDet(compound_coef=7, num_classes=num_classes, **kwargs)



import pandas as pd
import numpy as np
import openml
import math

from typing import SupportsIndex, Tuple, ClassVar
from pathlib import Path
import re

import lucid
from lucid._tensor import Tensor

from ._base import DatasetBase


__all__ = ["MNIST", "FashionMNIST"]


class MNIST(DatasetBase):
    OPENML_ID: ClassVar[int] = 554

    def __init__(
        self,
        root: str | Path,
        train: bool | None = True,
        download: bool | None = False,
        transform: lucid.nn.Module | None = None,
        target_transform: lucid.nn.Module | None = None,
        test_size: float = 0.2,
        to_tensor: bool = True,
        *,
        cache: bool = True,
        scale: float | None = None,
        resize: tuple[int, int] | None = None,
        normalize: tuple[tuple[float, ...], tuple[float, ...]] | None = None,
        cache_preprocessed: bool = True,
        preprocess_dtype: lucid.Numeric = lucid.Float16,
        preprocess_chunk_size: int = 4096,
    ) -> None:
        self.cache = cache
        self.scale = scale
        self.resize = resize
        self.normalize = normalize
        self.cache_preprocessed = cache_preprocessed
        self.preprocess_dtype = preprocess_dtype
        self.preprocess_chunk_size = preprocess_chunk_size

        super().__init__(
            root=root,
            train=train,
            download=download,
            transform=transform,
            target_transform=target_transform,
            test_size=test_size,
            to_tensor=to_tensor,
        )

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "MNIST.csv", index=False)

        except Exception as e:
            raise RuntimeError(f"Failed to download the MNIST dataset. Error: {e}")

    def _cache_key(self) -> str:
        parts: list[str] = []
        if self.scale is not None:
            parts.append(f"s{self.scale:g}")
        if self.resize is not None:
            parts.append(f"r{self.resize[0]}x{self.resize[1]}")
        if self.normalize is not None:
            mean, std = self.normalize
            parts.append("m" + ",".join(f"{v:g}" for v in mean))
            parts.append("v" + ",".join(f"{v:g}" for v in std))
        if not parts:
            return "raw"
        key = "_".join(parts)
        return re.sub(r"[^a-zA-Z0-9_,.x-]+", "_", key)

    def _raw_cache_path(self) -> Path:
        return self.root / "MNIST_int16.npz"

    def _proc_cache_path(self) -> Path:
        dtype_name = str(self.preprocess_dtype)
        return self.root / f"MNIST_{self._cache_key()}_{dtype_name}.npz"

    def _ensure_raw_cache(self) -> tuple[np.ndarray, np.ndarray]:
        raw_path = self._raw_cache_path()
        if self.cache and raw_path.exists():
            with np.load(raw_path) as npz:
                images = npz["images"]
                labels = npz["labels"]
            return images, labels

        csv_path = self.root / "MNIST.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"MNIST dataset CSV file not found at {csv_path}. Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int32)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 1, 28, 28).astype(np.int16)

        if self.cache:
            np.savez_compressed(raw_path, images=images, labels=labels)

        return images, labels

    def _maybe_preprocess_and_cache(
        self, images_int16: np.ndarray, labels_int32: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        if self.resize is None and self.scale is None and self.normalize is None:
            return images_int16.astype(np.float32), labels_int32

        proc_path = self._proc_cache_path()
        if self.cache and self.cache_preprocessed and proc_path.exists():
            with np.load(proc_path) as npz:
                images = npz["images"]
                labels = npz["labels"]
            return images, labels

        from lucid.transforms import Compose, Resize, Normalize

        class _Scale(lucid.nn.Module):
            def __init__(self, factor: float) -> None:
                super().__init__()
                self.factor = factor

            def forward(self, x: Tensor) -> Tensor:
                return x * self.factor

        transforms: list[lucid.nn.Module] = []
        if self.resize is not None:
            transforms.append(Resize(self.resize))
        if self.scale is not None:
            transforms.append(_Scale(self.scale))
        if self.normalize is not None:
            mean, std = self.normalize
            transforms.append(Normalize(mean=mean, std=std))

        transform = Compose(transforms)
        n = images_int16.shape[0]
        out_h, out_w = self.resize if self.resize is not None else (28, 28)

        out_dtype = np.float16 if self.preprocess_dtype == lucid.Float16 else np.float32
        out_images = np.empty((n, 1, out_h, out_w), dtype=out_dtype)

        for start in range(0, n, self.preprocess_chunk_size):
            end = min(start + self.preprocess_chunk_size, n)
            chunk = images_int16[start:end].astype(np.float32)
            x = lucid.to_tensor(chunk, dtype=lucid.Float32)
            x = transform(x)
            out_images[start:end] = x.numpy().astype(out_dtype, copy=False)

        if self.cache and self.cache_preprocessed:
            np.savez_compressed(proc_path, images=out_images, labels=labels_int32)

        return out_images, labels_int32

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        images, labels = self._ensure_raw_cache()
        images, labels = self._maybe_preprocess_and_cache(images, labels)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            if images.dtype == np.float16 and self.preprocess_dtype == lucid.Float16:
                images = lucid.to_tensor(images, dtype=lucid.Float16)
            else:
                images = lucid.to_tensor(images, dtype=lucid.Float32)
            labels = lucid.to_tensor(labels, dtype=lucid.Int32)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index]
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label


class FashionMNIST(DatasetBase):
    OPENML_ID: ClassVar[int] = 40996

    def _download(self) -> None:
        try:
            dataset = openml.datasets.get_dataset(self.OPENML_ID)
            df, _, _, _ = dataset.get_data(dataset_format="dataframe")
            df.to_csv(self.root / "FashionMNIST.csv", index=False)
        except Exception as e:
            raise RuntimeError(
                f"Failed to download the FashionMNIST dataset. Error: {e}"
            )

    def _load_data(self, split: str) -> Tuple[Tensor, Tensor]:
        csv_path = self.root / "FashionMNIST.csv"
        if not csv_path.exists():
            raise RuntimeError(
                f"FashionMNIST dataset CSV file not found at {csv_path}. "
                + "Use `download=True`."
            )

        df = pd.read_csv(csv_path)
        labels = df["class"].values.astype(np.int32)
        images = df.drop(columns=["class"]).values.astype(np.float32)
        images = images.reshape(-1, 1, 28, 28)

        train_size = int(math.floor(len(images) * (1 - self.test_size)))
        if split == "train":
            images, labels = images[:train_size], labels[:train_size]
        else:
            images, labels = images[train_size:], labels[train_size:]

        if self.to_tensor:
            images = lucid.to_tensor(images, dtype=lucid.Float32)
            labels = lucid.to_tensor(labels, dtype=lucid.Int32)

        return images, labels

    def __getitem__(self, index: SupportsIndex) -> Tuple[Tensor, Tensor]:
        image = self.data[index].reshape(-1, 1, 28, 28)
        label = self.targets[index]

        if self.transform:
            image = self.transform(image)
        if self.target_transform:
            label = self.target_transform(label)

        return image, label



from functools import partial
from types import ModuleType
from typing import TypeAlias
import itertools
import os

import numpy as np

from lucid._tensor import Tensor
from lucid._backend.core import (
    Operation,
    func_op,
    _FuncOpReturnType,
    _GradType,
)
from lucid._backend.metal import mx

from lucid.types import _NumPyArray, _MLXArray


_Array: TypeAlias = _NumPyArray | _MLXArray
_Shape: TypeAlias = tuple[int, ...]
_Stride: TypeAlias = tuple[int, ...]
_Padding: TypeAlias = tuple[int, ...]
_Dilation: TypeAlias = tuple[int, ...]


def _load_view_limit_bytes() -> int:
    env = os.getenv("LUCID_CONV_VIEW_LIMIT_MB")
    if env is None:
        return _default_view_limit_bytes()
    try:
        value = int(env)
    except ValueError:
        return _default_view_limit_bytes()
    return value * 1024 * 1024


def _sysconf_value(name: str) -> int | None:
    try:
        value = int(os.sysconf(name))
    except (ValueError, AttributeError, OSError):
        return None
    if value <= 0:
        return None
    return value


def _get_total_memory_bytes() -> int | None:
    page_size = _sysconf_value("SC_PAGE_SIZE") or _sysconf_value("SC_PAGESIZE")
    phys_pages = _sysconf_value("SC_PHYS_PAGES")
    if page_size and phys_pages:
        return page_size * phys_pages
    try:
        import ctypes

        class MEMORYSTATUSEX(ctypes.Structure):
            _fields_ = [
                ("dwLength", ctypes.c_ulong),
                ("dwMemoryLoad", ctypes.c_ulong),
                ("ullTotalPhys", ctypes.c_ulonglong),
                ("ullAvailPhys", ctypes.c_ulonglong),
                ("ullTotalPageFile", ctypes.c_ulonglong),
                ("ullAvailPageFile", ctypes.c_ulonglong),
                ("ullTotalVirtual", ctypes.c_ulonglong),
                ("ullAvailVirtual", ctypes.c_ulonglong),
                ("ullAvailExtendedVirtual", ctypes.c_ulonglong),
            ]

        stat = MEMORYSTATUSEX()
        stat.dwLength = ctypes.sizeof(MEMORYSTATUSEX)
        if ctypes.windll.kernel32.GlobalMemoryStatusEx(ctypes.byref(stat)):
            return int(stat.ullTotalPhys)

    except Exception:
        return None


def _round_to_step(value: int, step: int) -> int:
    return ((value + step // 2) // step) * step


def _default_view_limit_bytes() -> int:
    total = _get_total_memory_bytes()
    if not total:
        return 256 * 1024 * 1024

    mb = 1024 * 1024
    min_bytes = 64 * mb
    max_bytes = 1024 * mb
    step = 64 * mb

    target = (total * 15) // 1000
    target = max(min_bytes, min(max_bytes, target))
    target = _round_to_step(target, step)
    return max(min_bytes, min(max_bytes, target))


_CONV_VIEW_LIMIT_BYTES = _load_view_limit_bytes()


def get_conv_view_limit_mb() -> int:
    return int(_CONV_VIEW_LIMIT_BYTES // (1024 * 1024))


def _dtype_itemsize(data: _Array) -> int:
    dtype = getattr(data, "dtype", None)
    if dtype is None:
        return 0
    try:
        return int(np.dtype(dtype).itemsize)
    except TypeError:
        return int(getattr(dtype, "size", 0) or 0)


def _prod(shape: _Shape) -> int:
    total = 1
    for v in shape:
        total *= int(v)
    return total


def _view_exceeds_limit(data: _Array, out_dims: _Shape, kernel_size: _Shape) -> bool:
    if _CONV_VIEW_LIMIT_BYTES == 0:
        return True
    if _CONV_VIEW_LIMIT_BYTES < 0:
        return False
    itemsize = _dtype_itemsize(data)
    if itemsize == 0:
        return False

    view_elems = data.shape[0] * data.shape[1] * _prod(out_dims) * _prod(kernel_size)
    view_bytes = view_elems * itemsize

    return view_bytes > _CONV_VIEW_LIMIT_BYTES


def _to_tuple(value: int | tuple[int, ...] | list[int], dim: int, name: str) -> _Shape:
    if isinstance(value, int):
        return (value,) * dim

    if isinstance(value, (tuple, list)):
        if len(value) == 1:
            return (int(value[0]),) * dim
        if len(value) != dim:
            raise ValueError(f"{name} must have length {dim}, got {len(value)}.")
        return tuple(int(v) for v in value)

    raise TypeError(f"{name} must be int or sequence, got {type(value).__name__}.")


def _conv_out_dims(
    input_spatial: _Shape,
    kernel_size: _Shape,
    stride: _Stride,
    padding: _Padding,
    dilation: _Dilation,
) -> list[int]:
    out_dims = []
    for i in range(len(kernel_size)):
        eff = dilation[i] * (kernel_size[i] - 1) + 1
        o = (input_spatial[i] + 2 * padding[i] - eff) // stride[i] + 1
        if o <= 0:
            raise ValueError(f"Non-positive output dim for axis {i}: {o}")
        out_dims.append(o)

    return out_dims


def _validate_conv_shapes(input_: Tensor, weight: Tensor, groups: int) -> None:
    if input_.ndim != weight.ndim:
        raise ValueError("Input and weight must have the same number of dimensions.")
    if input_.ndim < 3:
        raise ValueError("Input and weight must have at least 3 dimensions.")
    if groups <= 0:
        raise ValueError("groups must be a positive integer.")

    C_in = input_.shape[1]
    C_out = weight.shape[0]
    C_in_g = weight.shape[1]

    if C_out % groups != 0 or C_in_g * groups != C_in:
        raise ValueError("Inconsistent channel/group configuration.")


def _pad_input(lib_: ModuleType, data: _Array, padding: _Padding) -> _Array:
    if not any(padding):
        return data

    pad_width = ((0, 0), (0, 0)) + tuple((p, p) for p in padding)
    return lib_.pad(data, pad_width)


def _as_strided(
    lib_: ModuleType, data: _Array, shape: _Shape, strides: _Shape
) -> _Array | None:
    if lib_ is np:
        return np.lib.stride_tricks.as_strided(data, shape=shape, strides=strides)

    as_strided = getattr(lib_, "as_strided", None)
    if as_strided is None:
        return None

    try:
        return as_strided(data, shape=shape, strides=strides)
    except TypeError:
        return as_strided(data, shape, strides)


def _make_input_view(
    lib_: ModuleType,
    data: _Array,
    out_dims: _Shape,
    kernel_size: _Shape,
    stride: _Stride,
    dilation: _Dilation,
) -> _Array | None:
    if not hasattr(data, "strides"):
        return None
    strides = data.strides
    if strides is None:
        return None

    spatial_strides = strides[2:]
    view_strides = (
        strides[0],
        strides[1],
        *[spatial_strides[i] * stride[i] for i in range(len(kernel_size))],
        *[spatial_strides[i] * dilation[i] for i in range(len(kernel_size))],
    )
    view_shape = (data.shape[0], data.shape[1], *out_dims, *kernel_size)

    return _as_strided(lib_, data, view_shape, view_strides)


def _conv_from_view(
    lib_: ModuleType, x_view: _Array, weight: _Array, out_dims: _Shape, groups: int
) -> _Array:
    D = len(out_dims)
    C_out = weight.shape[0]
    C_in_g = weight.shape[1]
    C_out_g = C_out // groups

    axes_x = [1] + list(range(2 + D, 2 + 2 * D))
    axes_w = [1] + list(range(2, 2 + D))
    perm = [0, D + 1] + list(range(1, D + 1))

    outputs = []
    for g in range(groups):
        x_g = x_view[:, g * C_in_g : (g + 1) * C_in_g, ...]
        w_g = weight[g * C_out_g : (g + 1) * C_out_g, ...]

        out = lib_.tensordot(x_g, w_g, axes=(axes_x, axes_w))
        out = lib_.transpose(out, axes=perm)
        outputs.append(out)

    if len(outputs) == 1:
        return outputs[0]

    return lib_.concatenate(outputs, axis=1)


def _conv_fallback(
    lib_: ModuleType,
    input_: _Array,
    weight: _Array,
    stride: _Stride,
    padding: _Padding,
    dilation: _Dilation,
    groups: int,
    out_dims: _Shape,
) -> _Array:
    D = len(out_dims)
    kernel_size = weight.shape[2:]
    C_out = weight.shape[0]
    C_in_g = weight.shape[1]
    C_out_g = C_out // groups

    x = _pad_input(lib_, input_, padding)

    outputs = []
    for g in range(groups):
        x_g = x[:, g * C_in_g : (g + 1) * C_in_g]
        w_g = weight[g * C_out_g : (g + 1) * C_out_g]

        out_g = None
        for k_idx in itertools.product(*[range(k) for k in kernel_size]):
            slices = [slice(None), slice(None)]

            for d in range(D):
                start = k_idx[d] * dilation[d]
                end = start + stride[d] * out_dims[d]
                slices.append(slice(start, end, stride[d]))

            x_slice = x_g[tuple(slices)]
            w_slice = w_g[(slice(None), slice(None)) + k_idx]

            contrib = lib_.tensordot(x_slice, w_slice, axes=([1], [1]))
            perm = [0, contrib.ndim - 1] + list(range(1, contrib.ndim - 1))

            contrib = lib_.transpose(contrib, axes=perm)
            out_g = contrib if out_g is None else out_g + contrib

        outputs.append(out_g)

    if len(outputs) == 1:
        return outputs[0]

    return lib_.concatenate(outputs, axis=1)


def _conv_forward(
    lib_: ModuleType,
    input_: _Array,
    weight: _Array,
    stride: _Stride,
    padding: _Padding,
    dilation: _Dilation,
    groups: int,
) -> _Array:
    input_spatial = input_.shape[2:]
    kernel_size = weight.shape[2:]
    out_dims = tuple(
        _conv_out_dims(input_spatial, kernel_size, stride, padding, dilation)
    )

    if _view_exceeds_limit(input_, out_dims, kernel_size):
        return _conv_fallback(
            lib_, input_, weight, stride, padding, dilation, groups, out_dims
        )

    x = _pad_input(lib_, input_, padding)
    x_view = _make_input_view(lib_, x, out_dims, kernel_size, stride, dilation)
    if x_view is None:
        return _conv_fallback(
            lib_, input_, weight, stride, padding, dilation, groups, out_dims
        )

    return _conv_from_view(lib_, x_view, weight, out_dims, groups)


def _conv_backward_weight(
    lib_: ModuleType,
    grad_out: _Array,
    x_pad: _Array,
    weight: _Array,
    stride: _Stride,
    dilation: _Dilation,
    groups: int,
) -> _Array:
    weight_shape = weight.shape
    D = len(weight_shape) - 2
    out_dims = grad_out.shape[2:]
    kernel_size = weight.shape[2:]
    C_out = weight_shape[0]
    C_in_g = weight_shape[1]
    C_out_g = C_out // groups

    x_view = _make_input_view(lib_, x_pad, out_dims, kernel_size, stride, dilation)
    if x_view is not None and _view_exceeds_limit(x_pad, out_dims, kernel_size):
        x_view = None
    axes_out = [0] + list(range(2, 2 + D))
    axes_x = [0] + list(range(2, 2 + D))

    grad_parts = []
    for g in range(groups):
        grad_out_g = grad_out[:, g * C_out_g : (g + 1) * C_out_g, ...]

        if x_view is None:
            x_g = x_pad[:, g * C_in_g : (g + 1) * C_in_g]
            grad_w = lib_.zeros((C_out_g, C_in_g, *kernel_size), dtype=weight.dtype)

            for k_idx in itertools.product(*[range(k) for k in kernel_size]):
                slices = [slice(None), slice(None)]

                for d in range(D):
                    start = k_idx[d] * dilation[d]
                    end = start + stride[d] * out_dims[d]
                    slices.append(slice(start, end, stride[d]))

                x_slice = x_g[tuple(slices)]
                w_grad = lib_.tensordot(grad_out_g, x_slice, axes=(axes_out, axes_x))

                if lib_ is np:
                    grad_w[(slice(None), slice(None)) + k_idx] = w_grad
                else:
                    grad_w = grad_w.at[(slice(None), slice(None)) + k_idx].add(w_grad)
            grad_parts.append(grad_w)

        else:
            x_view_g = x_view[:, g * C_in_g : (g + 1) * C_in_g, ...]
            grad_w = lib_.tensordot(grad_out_g, x_view_g, axes=(axes_out, axes_x))
            grad_parts.append(grad_w)

    if len(grad_parts) == 1:
        return grad_parts[0]

    return lib_.concatenate(grad_parts, axis=0)


def _conv_backward_input(
    lib_: ModuleType,
    grad_out: _Array,
    weight: _Array,
    x_pad: _Array,
    stride: _Stride,
    padding: _Padding,
    dilation: _Dilation,
    groups: int,
) -> _Array:
    kernel_size = weight.shape[2:]
    D = len(kernel_size)
    out_dims = grad_out.shape[2:]

    C_out = weight.shape[0]
    C_in_g = weight.shape[1]
    C_out_g = C_out // groups

    grad_input = lib_.zeros_like(x_pad)

    for g in range(groups):
        grad_out_g = grad_out[:, g * C_out_g : (g + 1) * C_out_g, ...]
        w_g = weight[g * C_out_g : (g + 1) * C_out_g]
        ch_slice = slice(g * C_in_g, (g + 1) * C_in_g)

        for k_idx in itertools.product(*[range(k) for k in kernel_size]):
            w_slice = w_g[(slice(None), slice(None)) + k_idx]
            contrib = lib_.tensordot(grad_out_g, w_slice, axes=([1], [0]))

            perm = [0, contrib.ndim - 1] + list(range(1, contrib.ndim - 1))
            contrib = lib_.transpose(contrib, axes=perm)

            slices = [slice(None), ch_slice]
            for d in range(D):
                start = k_idx[d] * dilation[d]
                end = start + stride[d] * out_dims[d]
                slices.append(slice(start, end, stride[d]))

            if lib_ is np:
                grad_input[tuple(slices)] += contrib
            else:
                grad_input = grad_input.at[tuple(slices)].add(contrib)

    if any(padding):
        crop = [slice(None), slice(None)]
        for p in padding:
            end = -p if p != 0 else None
            crop.append(slice(p, end))
        return grad_input[tuple(crop)]

    return grad_input


class conv_nd_kernel(Operation):
    def __init__(
        self,
        stride: int | tuple[int, ...] | list[int],
        padding: int | tuple[int, ...] | list[int],
        dilation: int | tuple[int, ...] | list[int],
        groups: int,
    ) -> None:
        super().__init__()
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups

        self._stride: _Stride | None = None
        self._padding: _Padding | None = None
        self._dilation: _Dilation | None = None

    def _normalize(self, weight: Tensor) -> tuple[_Stride, _Padding, _Dilation]:
        D = weight.ndim - 2
        stride = _to_tuple(self.stride, D, "stride")
        padding = _to_tuple(self.padding, D, "padding")
        dilation = _to_tuple(self.dilation, D, "dilation")

        self._stride = stride
        self._padding = padding
        self._dilation = dilation

        return stride, padding, dilation

    @func_op(n_in=2, n_ret=1)
    def cpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        _validate_conv_shapes(a, b, self.groups)
        stride, padding, dilation = self._normalize(b)
        out = _conv_forward(np, a.data, b.data, stride, padding, dilation, self.groups)

        self.result = Tensor(out)
        return self.result, partial(self.__grad__, a=a, b=b, lib_=np)

    @func_op(n_in=2, n_ret=1, device="gpu")
    def gpu(self, a: Tensor, b: Tensor) -> _FuncOpReturnType:
        _validate_conv_shapes(a, b, self.groups)
        stride, padding, dilation = self._normalize(b)
        out = _conv_forward(mx, a.data, b.data, stride, padding, dilation, self.groups)

        self.result = Tensor(out)
        return self.result, partial(self.__grad__, a=a, b=b, lib_=mx)

    def __grad__(self, a: Tensor, b: Tensor, lib_: ModuleType) -> _GradType:
        stride = self._stride
        padding = self._padding
        dilation = self._dilation

        if stride is None or padding is None or dilation is None:
            raise RuntimeError("conv_nd backward called before forward.")

        x_pad = _pad_input(lib_, a.data, padding)
        grad_out = self.result.grad

        grad_input = _conv_backward_input(
            lib_, grad_out, b.data, x_pad, stride, padding, dilation, self.groups
        )
        grad_weight = _conv_backward_weight(
            lib_, grad_out, x_pad, b.data, stride, dilation, self.groups
        )

        return grad_input, grad_weight

    def __flops__(self, a: Tensor, b: Tensor) -> int:
        stride = self._stride
        padding = self._padding
        dilation = self._dilation
        if stride is None or padding is None or dilation is None:
            stride, padding, dilation = self._normalize(b)

        N = int(a.shape[0])
        C_out = int(b.shape[0])
        C_in_g = int(b.shape[1])
        kernel_size = tuple(int(v) for v in b.shape[2:])
        out_dims = _conv_out_dims(
            tuple(int(v) for v in a.shape[2:]), kernel_size, stride, padding, dilation
        )

        macs_per_out = C_in_g * _prod(kernel_size)
        out_elems = N * C_out * _prod(tuple(out_dims))
        return out_elems * macs_per_out



from typing import Iterable, Sequence
import weakref

import numpy as np

from lucid._backend.metal import mx
from lucid.error import BackwardError
from lucid.types import _MLXArray, _NumPyArray, _TensorLike, _Scalar, _Gradient


__all__ = ["grad", "backward"]


def _as_tuple(value: _TensorLike | Sequence[_TensorLike]) -> tuple[_TensorLike, ...]:
    if isinstance(value, (tuple, list)):
        return tuple(value)
    return (value,)


def _coerce_grad_output(
    output: _TensorLike, grad_output: _TensorLike | _Gradient | _Scalar
) -> _Gradient:
    if grad_output is None:
        if output.is_cpu():
            return np.ones_like(output.data)
        return mx.ones_like(output.data)

    if isinstance(grad_output, _TensorLike):
        grad_output = grad_output.data

    if isinstance(grad_output, _NumPyArray):
        if output.is_gpu():
            grad_output = mx.array(grad_output)
        return grad_output

    if isinstance(grad_output, _MLXArray):
        if output.is_cpu():
            grad_output = np.array(grad_output)
        return grad_output

    if output.is_cpu():
        return np.ones_like(output.data) * grad_output
    return mx.ones_like(output.data) * grad_output


def grad(
    outputs: _TensorLike | Iterable[_TensorLike],
    inputs: _TensorLike | Iterable[_TensorLike],
    grad_outputs: _TensorLike | Iterable[_TensorLike] | Iterable[_Scalar] | None = None,
    retain_graph: bool = False,
    allow_unused: bool = False,
) -> tuple[_Gradient, ...] | _Gradient:
    out_tensors = _as_tuple(outputs)
    in_tensors = _as_tuple(inputs)

    if grad_outputs is None:
        grad_outs = (None,) * len(out_tensors)
    else:
        grad_outs = _as_tuple(grad_outputs)
        if len(grad_outs) != len(out_tensors):
            raise ValueError("grad_outputs length must match outputs length.")

    for tensor in out_tensors:
        if not isinstance(tensor, _TensorLike):
            raise TypeError("All outputs must be _TensorLike instances.")

    for tensor in in_tensors:
        if not isinstance(tensor, _TensorLike):
            raise TypeError("All inputs must be _TensorLike instances.")

    prev_grads = {tensor: tensor.grad for tensor in in_tensors}
    prev_out_grads = {tensor: tensor.grad for tensor in out_tensors}
    prev_keep = {tensor: tensor.keep_grad for tensor in in_tensors}

    for tensor in in_tensors:
        tensor.grad = None
        tensor.keep_grad = True

    try:
        for i, (output, grad_output) in enumerate(zip(out_tensors, grad_outs)):
            if not output.requires_grad:
                if allow_unused:
                    continue
                raise RuntimeError("All outputs must require gradients.")

            coerced = _coerce_grad_output(output, grad_output)
            if coerced.shape != output.shape:
                raise ValueError("grad_output shape must match output shape.")

            output.grad = coerced
            output.backward(
                retain_grad=True,
                retain_graph=(i < len(out_tensors) - 1) or retain_graph,
            )

        grads = tuple(tensor.grad for tensor in in_tensors)
        if not allow_unused and any(grad is None for grad in grads):
            raise RuntimeError(
                "Some inputs did not receive gradients. Set allow_unused=True."
            )

        if len(grads) == 1:
            return grads[0]
        return grads
    finally:
        for tensor, grad in prev_out_grads.items():
            tensor.grad = grad
        for tensor in in_tensors:
            tensor.grad = prev_grads[tensor]
            tensor.keep_grad = prev_keep[tensor]


def backward(
    tensor: _TensorLike, retain_grad: bool = False, retain_graph: bool = False
) -> None:
    if tensor.grad is None:
        tensor.grad = (
            np.ones_like(tensor.data) if tensor.is_cpu() else mx.ones_like(tensor.data)
        )

    visited = set()
    topo_order: list[_TensorLike] = []
    stack = [tensor]
    ops_to_clear = set()

    while stack:
        node = stack[-1]
        if node in visited:
            stack.pop()
            topo_order.append(node)
            continue

        visited.add(node)
        for parent in node._prev:
            if parent not in visited:
                stack.append(parent)

    from lucid._fusion import ENABLE_FUSION

    if ENABLE_FUSION and tensor.is_cpu():
        _try_backward_fusion(topo_order)

    for node in reversed(topo_order):
        try:
            node._backward_op()
        except Exception as e:
            raise BackwardError(shape=node.shape, op=node._op) from e

        for hook in node._backward_hooks:
            hook(node, node.grad)

        if node._op is not None:
            ops_to_clear.add(node._op)

        if not (node.is_leaf or retain_grad or node.keep_grad):
            node.grad = None

    if not retain_graph:
        for node in topo_order:
            node.clear_node()
        for op in ops_to_clear:
            try:
                op.clear()
            except Exception:
                try:
                    op.result = None
                except Exception:
                    pass


def _try_backward_fusion(topo_order: list[_TensorLike]) -> None:
    from lucid._fusion import match_fusion_table

    consumer_of: dict[int, _TensorLike] = {}
    multi_consumer: set[int] = set()

    for consumer in topo_order:
        for parent in consumer._prev:
            pid = id(parent)
            if pid in multi_consumer:
                continue

            prev_consumer = consumer_of.get(pid)
            if prev_consumer is None:
                consumer_of[pid] = consumer
            else:
                multi_consumer.add(pid)
                consumer_of.pop(pid, None)

    if not consumer_of:
        return

    for pid, v in list(consumer_of.items()):
        p = next((t for t in v._prev if id(t) == pid), None)
        if p is None:
            continue
        if p._op is None or v._op is None:
            continue

        fused_backward_op = match_fusion_table(p._op, v._op)
        if fused_backward_op is None:
            continue
        if v.size < fused_backward_op.heuristic_thresh:
            continue

        if len(v._prev) != 1 or v._prev[0] is not p:
            continue

        p_parents = tuple(p._prev)
        v._prev.remove(p)
        v._prev.extend(p_parents)
        p.clear_node(clear_op=False)

        v._backward_op.override_tensor_refs(
            new_tensor_refs=tuple(weakref.ref(t) for t in v._prev),
            new_versions=tuple(t._version for t in v._prev),
        )
        v._backward_op.override_grad_func(
            fused_backward_op.get_fused_grad_func(
                inputs=p_parents, results=v, device=v.device
            )
        )



from typing import Callable, Iterator, Optional, Self, SupportsIndex, Any, overload
from types import NoneType
from collections import deque

import numpy as np
import weakref

import lucid
from lucid import types
from lucid.types import (
    _ArrayOrScalar,
    _NumPyArray,
    _MLXArray,
    _Scalar,
    _DeviceType,
    _BuiltinNumeric,
    Numeric,
)

from lucid._tensor.base import _TensorBase, _TensorInplace
from lucid._backend.core import BackwardOperation, Operation, noop
from lucid._backend.metal import mx, parse_mlx_indexing, check_metal_availability


_HookType = Callable[["Tensor", _NumPyArray | _MLXArray], None]

_dtype_map = {int: types.Int64, float: types.Float64, complex: types.Complex64}


class Tensor(_TensorBase, _TensorInplace):
    def __init__(
        self,
        data: _ArrayOrScalar | _MLXArray,
        requires_grad: bool = False,
        keep_grad: bool = False,
        dtype: _BuiltinNumeric | Numeric | None = None,
        device: _DeviceType = "cpu",
    ) -> None:
        self._is_free = False
        self._is_bool_tensor = False

        if dtype is bool:
            self._is_bool_tensor = True
            dtype = None
        if dtype in {int, float, complex}:
            dtype = _dtype_map[dtype]

        dtype: Numeric | None
        assert isinstance(dtype, (Numeric, NoneType)), type(dtype)

        if dtype is not None and dtype.is_bit_free:
            raise TypeError(
                f"Implicit dtypes is not supported for Tensor instantiation."
            )
        if device not in {"cpu", "gpu"}:
            raise lucid.UnknownDeviceError(device)

        self.data: _NumPyArray | _MLXArray
        if not isinstance(data, (_NumPyArray, _MLXArray)):
            self.data = np.array(data, dtype=dtype.cpu if dtype is not None else dtype)

            if isinstance(data, (_Scalar, list, tuple)):
                self._is_free = True

        elif isinstance(data, _NumPyArray):
            self.data = data
            if dtype is not None and data.dtype != dtype.cpu:
                self.data = data.astype(dtype.cpu)

        elif isinstance(data, _MLXArray):
            check_metal_availability()
            device = "gpu"
            self.data = data
            if dtype is not None and data.dtype != dtype.gpu:
                self.data = data.astype(dtype.gpu)

        if device == "gpu" and isinstance(self.data, _NumPyArray):
            check_metal_availability()
            self.data = mx.array(self.data)

        if "bool" in str(self.data.dtype):
            self._is_bool_tensor = True
        else:
            if self._is_bool_tensor:
                self.data = self.data.astype(bool if device == "cpu" else mx.bool_)

        self._op: Operation | None = None
        self._backward_op: BackwardOperation = noop
        self._prev: list[Tensor] = []
        self._backward_hooks: list[_HookType] = []
        self._version: int = 0

        self.grad: Optional[_NumPyArray | _MLXArray] = None
        self.device = device
        self.requires_grad = requires_grad and lucid.grad_enabled()
        self.keep_grad = keep_grad

        self.dtype: Numeric | bool
        if self._is_bool_tensor:
            self.dtype = bool
        else:
            self.dtype = (
                dtype if dtype is not None else types.to_numeric_type(self.data.dtype)
            )

    @property
    def is_leaf(self) -> bool:
        return self.requires_grad and len(self._prev) == 0

    @property
    def is_free(self) -> bool:
        return self._is_free

    @classmethod
    def is_all_free(cls, *tensors: Self) -> bool:
        return all(t.is_free for t in tensors)

    @classmethod
    def copy_data(cls, data: _NumPyArray | _MLXArray) -> _NumPyArray | _MLXArray:
        if isinstance(data, _NumPyArray):
            return data.copy()
        elif isinstance(data, _MLXArray):
            return mx.array(data)
        else:
            raise TypeError(f"Unexpected type: '{type(data).__name__}'")

    @classmethod
    def copy_grad(cls, grad: _NumPyArray | _MLXArray) -> _NumPyArray | _MLXArray:
        return cls.copy_data(data=grad)

    def new_tensor(self) -> Tensor:
        return Tensor(
            self.data, self.requires_grad, self.keep_grad, self.dtype, self.device
        )

    def free(self) -> Self:
        self._is_free = True
        return self

    def eval(self) -> Self:
        if self.is_gpu():
            mx.eval(self.data)
            stopped = mx.stop_gradient(self.data)
            if stopped is not None:
                self.data = stopped
        return self

    def backward(self, retain_grad: bool = False, retain_graph: bool = False) -> None:
        import lucid.autograd as autograd

        autograd.backward(self, retain_grad=retain_grad, retain_graph=retain_graph)

    def register_hook(self, hook: _HookType) -> Callable:
        self._backward_hooks.append(hook)
        return lambda: self._backward_hooks.remove(hook)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.data.shape

    @property
    def ndim(self) -> int:
        return self.data.ndim

    @property
    def size(self) -> int:
        return self.data.size

    @property
    def flops(self) -> int:
        total = 0
        queue = deque([self])
        visited = set()

        while queue:
            cur = queue.popleft()
            visited.add(cur)
            if cur._op is not None:
                total += cur._op.flops

            for next in cur._prev:
                if next not in visited:
                    queue.appendleft(next)

        return total

    def item(self) -> _Scalar | bool:
        if self.size > 1:
            raise ValueError(
                "Tensor must be 0-dimensional(scalar) to pop its item. "
                "Use `tensor.data` to access the data directly.",
            )
        item = self.data[..., 0] if self.ndim > 0 else self.data
        if self.dtype is bool:
            return bool(item)

        if item % 1 == 0:
            return int(item)
        else:
            return float(item)

    def tolist(self) -> list:
        return self.data.tolist()

    def numpy(self) -> _NumPyArray:
        return np.array(self.data)

    def mlx(self) -> _MLXArray:
        return mx.array(self.data)

    def detach(self) -> Self:
        data_copy = Tensor.copy_data(self.data)
        return Tensor(
            data_copy, requires_grad=False, dtype=self.dtype, device=self.device
        )

    def zero(self) -> None:
        self._version += 1
        if self.is_cpu():
            self.data = np.zeros_like(self.data)
        else:
            self.data = mx.zeros_like(self.data)

    def zero_grad(self) -> None:
        self.grad = None

    def clear_node(self, clear_op: bool = True) -> None:
        self._prev = []
        self._backward_op = noop
        if clear_op:
            self._op = None

    def astype(self, dtype: type | Numeric) -> Self:
        new_dtype = dtype
        if isinstance(dtype, Numeric):
            self._is_bool_tensor = False
            if dtype.is_bit_free:
                new_dtype = dtype.auto_parse(self.data.dtype, device=self.device)
            else:
                new_dtype = dtype.parse(device=self.device)

        elif dtype is bool:
            self._is_bool_tensor = True
            if self.is_gpu():
                new_dtype = mx.bool_

        self.data = self.data.astype(new_dtype)
        if self._is_bool_tensor:
            self.dtype = bool
        else:
            self.dtype = types.to_numeric_type(self.data.dtype)

        self._version += 1
        return self

    @overload
    def to(self, device: _DeviceType) -> Self: ...

    @overload
    def to(self, dtype: type | Numeric) -> Self: ...

    def to(self, device_or_dtype: _DeviceType | type | Numeric) -> Self:
        if isinstance(device_or_dtype, str):
            device = device_or_dtype
            if self.device == device:
                return self

            if device == "cpu":
                self.data = np.array(self.data)
                if self.grad is not None:
                    self.grad = np.array(self.grad)

            elif device == "gpu":
                check_metal_availability()
                self.data = mx.array(self.data)
                if self.grad is not None:
                    self.grad = mx.array(self.grad)

            else:
                raise lucid.UnknownDeviceError(device)

            self.device = device
            return self
        else:
            dtype = device_or_dtype
            return self.astype(dtype)

    def is_cpu(self) -> bool:
        return self.device == "cpu"

    def is_gpu(self) -> bool:
        return self.device == "gpu"

    def __getitem__(self, idx: SupportsIndex | Self) -> Self:
        new_idx = idx
        if isinstance(idx, Tensor):
            new_idx = idx.data
            if self.is_gpu() and idx.is_free:
                new_idx = mx.array(idx.data)

        if isinstance(idx, tuple):
            new_idx = tuple()
            for id in idx:
                if isinstance(id, Tensor):
                    if self.is_gpu() and id.is_free:
                        id = mx.array(id.data)
                    else:
                        id = id.data
                new_idx += (id,)

        if self.is_gpu():
            new_idx = parse_mlx_indexing(new_idx)
        else:
            if isinstance(idx, Tensor) and idx.is_gpu():
                raise lucid.DeviceMismatchError(to="gpu", from_="cpu")

        sliced_data = self.data[new_idx]
        new_tensor = Tensor(
            sliced_data, self.requires_grad, self.keep_grad, self.dtype, self.device
        )

        def _grad_func() -> None:
            if self.grad is None:
                self.grad = (
                    np.zeros_like(self.data)
                    if self.is_cpu()
                    else mx.zeros_like(self.data)
                )
            if new_tensor.grad is None:
                return
            new_grad = lucid._match_grad_shape(
                self.data[new_idx], new_tensor.grad, device=self.device
            )
            lucid._set_tensor_grad(self, new_grad, at=new_idx)

        if self.requires_grad and lucid.grad_enabled():
            new_tensor._prev = [self]
            new_tensor._backward_op = BackwardOperation(
                forward_op_ref=None,
                grad_func=None,
                tensor_refs=(weakref.ref(self),),
                device=self.device,
                versions=(self._version,),
                custom_closure=_grad_func,
            )

        if self.is_free:
            new_tensor.free()

        return new_tensor

    def __setitem__(
        self, idx: SupportsIndex | Self, value: Self | _ArrayOrScalar
    ) -> None:
        if self.requires_grad:
            raise RuntimeError(
                "Cannot perform in-place item setting on a "
                + "Tensor that requires gradients. "
            )

        new_idx = idx
        if isinstance(idx, Tensor):
            new_idx = idx.data
        elif isinstance(idx, tuple):
            new_idx = tuple()
            for id in idx:
                if isinstance(id, Tensor):
                    id = id.data
                new_idx += (id,)

        if self.is_gpu():
            new_idx = parse_mlx_indexing(new_idx)

        data: _ArrayOrScalar
        if isinstance(value, Tensor):
            if self.device != value.device:
                if value.is_free:
                    data = value.to(self.device).data
                else:
                    raise lucid.DeviceMismatchError(to=self.device, from_=value.device)
            data = value.data
        else:
            data = value
        self.data[new_idx] = data

    def __len__(self) -> int:
        if self.ndim == 0:
            return self.size
        else:
            return self.shape[0]

    def __iter__(self) -> Iterator[Self]:
        for i in range(self.shape[0]):
            yield self[i]

    def __repr__(self) -> str:
        return f"Tensor({self.data}, grad={self.grad}, device={self.device})"

    def __str__(self) -> str:
        return str(self.data)

    def __hash__(self) -> int:
        return hash(id(self))

    def __deepcopy__(self, *args: Any) -> Self:
        cls = self.__class__
        copied_data = Tensor.copy_data(self.data)

        if cls is Tensor:
            new = Tensor(
                copied_data, self.requires_grad, self.keep_grad, self.dtype, self.device
            )
        else:
            base = Tensor(copied_data, dtype=self.dtype, device=self.device)
            new = cls(base)

        if self.grad is not None and (
            self.keep_grad or getattr(new, "keep_grad", False)
        ):
            new.grad = Tensor.copy_grad(self.grad)
        else:
            new.grad = None

        new._op = None
        new._backward_op = noop
        new._prev = []
        new._backward_hooks = []

        new._is_free = self._is_free
        new._is_bool_tensor = self._is_bool_tensor

        return new

    def __bool__(self) -> bool:
        if self.data.size != 1:
            raise RuntimeError(
                "Boolean value of Tensor with more than one value is ambiguous. "
                "Use `tensor.any()` or `tensor.all()` instead."
            )
        return bool(self.data.item())

    def any(self, axis: int | None = None, keepdims: bool = False) -> bool | Self:
        if self.is_cpu():
            result = np.any(self.data, axis=axis, keepdims=keepdims)
            return bool(result) if axis is None else Tensor(result, device="cpu")
        else:
            mx.eval(self.data)
            result = mx.any(self.data, axis=axis, keepdims=keepdims)
            return bool(result.item()) if axis is None else Tensor(result, device="gpu")

    def all(self, axis=None, keepdims=False) -> bool | Self:
        if self.is_cpu():
            result = np.all(self.data, axis=axis, keepdims=keepdims)
            return bool(result) if axis is None else Tensor(result, device="cpu")
        else:
            mx.eval(self.data)
            result = mx.all(self.data, axis=axis, keepdims=keepdims)
            return bool(result.item()) if axis is None else Tensor(result, device="gpu")

    def char(self) -> Self:
        return self.astype(lucid.Char)

    def short(self) -> Self:
        return self.astype(lucid.Short)

    def int(self) -> Self:
        return self.astype(lucid.Int)

    def long(self) -> Self:
        return self.astype(lucid.Long)

    def half(self) -> Self:
        return self.astype(lucid.Half)

    def float(self) -> Self:
        return self.astype(lucid.Float)

    def double(self) -> Self:
        return self.astype(lucid.Double)

    def bool(self) -> Self:
        return self.astype(bool)



import functools
from types import ModuleType

import numpy as np

from lucid._backend.core import Operation, func_op, _FuncOpReturnType, _GradType
from lucid._backend.metal import mx
from lucid._tensor import Tensor
from lucid.types import _DeviceType


def _norm_axis(axis: int, ndim: int) -> int:
    return axis if axis >= 0 else axis + ndim


class softmax_kernel(Operation):
    def __init__(self, axis: int = -1) -> None:
        super().__init__()
        self.axis = axis
        self._axis = None
        self._y = None

    def clear(self) -> None:
        super().clear()
        self._axis = None
        self._y = None

    @func_op(n_in=1, n_ret=1, device="cpu")
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=np, device="cpu")

    @func_op(n_in=1, n_ret=1, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=mx, device="gpu")

    def _forward(
        self, a: Tensor, lib_: ModuleType, device: _DeviceType
    ) -> _FuncOpReturnType:
        axis = _norm_axis(self.axis, a.ndim)
        max_val = lib_.max(a.data, axis=axis, keepdims=True)
        exp_x = lib_.exp(a.data - max_val)
        sum_exp = lib_.sum(exp_x, axis=axis, keepdims=True)
        y = exp_x / sum_exp

        self._axis = axis
        self._y = y

        self.result = Tensor(y, device=device)
        return self.result, functools.partial(self.__grad__, lib_=lib_)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None:
            raise RuntimeError("softmax backward called before forward.")
        if self._y is None or self._axis is None:
            raise RuntimeError("softmax cached data missing.")

        dy = self.result.grad
        y = self._y
        axis = self._axis

        dot = lib_.sum(dy * y, axis=axis, keepdims=True)
        dx = y * (dy - dot)
        return dx


class sigmoid_kernel(Operation):
    def __init__(self) -> None:
        super().__init__()
        self._y = None

    def clear(self) -> None:
        super().clear()
        self._y = None

    @func_op(n_in=1, n_ret=1, device="cpu")
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=np, device="cpu")

    @func_op(n_in=1, n_ret=1, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=mx, device="gpu")

    def _forward(
        self, a: Tensor, lib_: ModuleType, device: _DeviceType
    ) -> _FuncOpReturnType:
        y = 1.0 / (1.0 + lib_.exp(-a.data))
        self._y = y
        self.result = Tensor(y, device=device)
        return self.result, functools.partial(self.__grad__)

    def __grad__(self) -> _GradType:
        if self.result is None or self.result.grad is None or self._y is None:
            raise RuntimeError("sigmoid backward called before forward.")

        dy = self.result.grad
        y = self._y

        dx = dy * y * (1 - y)
        return dx


class gelu_kernel(Operation):
    def __init__(self) -> None:
        super().__init__()
        self._x = None

    def clear(self) -> None:
        super().clear()
        self._x = None

    @func_op(n_in=1, n_ret=1, device="cpu")
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=np, device="cpu")

    @func_op(n_in=1, n_ret=1, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=mx, device="gpu")

    def _forward(
        self, a: Tensor, lib_: ModuleType, device: _DeviceType
    ) -> _FuncOpReturnType:
        self._x = a.data
        c = lib_.sqrt(2.0 / lib_.pi)
        y = 0.5 * a.data * (1.0 + lib_.tanh(c * (a.data + 0.044715 * (a.data**3))))

        self.result = Tensor(y, device=device)
        return self.result, functools.partial(self.__grad__, lib_=lib_)

    def __grad__(self, lib_: ModuleType) -> _GradType:
        if self.result is None or self.result.grad is None or self._x is None:
            raise RuntimeError("gelu backward called before forward.")

        x = self._x
        dy = self.result.grad
        c = lib_.sqrt(2.0 / lib_.pi)
        t = c * (x + 0.044715 * x**3)
        dt = c * (1 + 3 * 0.044715 * x**2)
        sech2 = 1.0 / lib_.cosh(t) ** 2

        dx = 0.5 * (1 + lib_.tanh(t)) + 0.5 * x * sech2 * dt
        return dy * dx


class silu_kernel(Operation):
    def __init__(self) -> None:
        super().__init__()
        self._x = None
        self._sig = None

    def clear(self) -> None:
        super().clear()
        self._x = None
        self._sig = None

    @func_op(n_in=1, n_ret=1, device="cpu")
    def cpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=np, device="cpu")

    @func_op(n_in=1, n_ret=1, device="gpu")
    def gpu(self, a: Tensor) -> _FuncOpReturnType:
        return self._forward(a, lib_=mx, device="gpu")

    def _forward(
        self, a: Tensor, lib_: ModuleType, device: _DeviceType
    ) -> _FuncOpReturnType:
        self._x = a.data
        sig = 1.0 / (1.0 + lib_.exp(-a.data))
        self._sig = sig
        y = a.data * sig

        self.result = Tensor(y, device=device)
        return self.result, functools.partial(self.__grad__)

    def __grad__(self) -> _GradType:
        if (
            self.result is None
            or self.result.grad is None
            or self._x is None
            or self._sig is None
        ):
            raise RuntimeError("silu backward called before forward.")

        dy = self.result.grad
        sig = self._sig
        x = self._x

        dx = dy * (sig + x * sig * (1 - sig))
        return dx



from typing import Iterable

import lucid

from lucid._tensor import Tensor
from lucid.types import _Scalar


__all__ = ["grad_norm", "clip_grad_norm", "clip_grad_value"]


def _as_iter(parameters: Iterable[Tensor] | Tensor) -> list[Tensor]:
    if isinstance(parameters, Tensor):
        return [parameters]
    return list(parameters)


def grad_norm(parameters: Iterable[Tensor] | Tensor, norm_type: int = 2) -> Tensor:
    parameters = _as_iter(parameters)
    device = parameters[0].device

    params: list[Tensor] = [p for p in parameters if p.grad is not None]
    if not params:
        return Tensor(0.0, device=device)

    norm_pow_sum = 0.0
    for p in params:
        param_norm = lucid.linalg.norm(lucid.ravel(p.grad), ord=norm_type).item()
        norm_pow_sum += param_norm**norm_type

    total_norm = norm_pow_sum ** (1.0 / norm_type)
    return Tensor(total_norm, device=device)


def clip_grad_norm(
    parameters: Iterable[Tensor] | Tensor,
    max_norm: _Scalar,
    norm_type: int = 2,
    eps: float = 1e-7,
) -> float:
    params: list[Tensor] = [p for p in _as_iter(parameters) if p.grad is not None]
    total_norm = grad_norm(params, norm_type=norm_type)

    clip_coef = float(max_norm) / (total_norm.item() + eps)
    if clip_coef < 1.0:
        for p in params:
            p.grad = p.grad * clip_coef

    return total_norm


def clip_grad_value(parameters: Iterable[Tensor] | Tensor, clip_value: _Scalar) -> None:
    params = [p for p in _as_iter(parameters) if p.grad is not None]
    if not params:
        return

    lo, hi = -float(clip_value), float(clip_value)
    for p in params:
        g_clip = lucid.clip(p.grad, lo, hi).data
        p.grad = g_clip
